/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package pkg3dtestapplication;

import java.util.ArrayList;
import javafx.event.EventHandler;
import javafx.geometry.Point2D;
import javafx.geometry.Point3D;
import javafx.scene.Group;
import javafx.scene.input.MouseEvent;
import javafx.scene.input.TouchEvent;
import javafx.scene.input.TouchPoint;
import javafx.scene.input.ZoomEvent;
import javafx.scene.paint.Color;
import javafx.scene.shape.Rectangle;


/**
 *
 * @author henri
 */
public class BoxModelWidget extends Xform {

    private final static int START = 0;
    private final static int SCALE_LOOPS_H = 1;
    private final static int EXTRUDE = 2;
    private final static int DRAW = 3;
    private final static int SCALE_LOOPS_V = 4;

    private int state;

    private float x, y, width, height;
    private int numEdgeTouchpoints = 0;
    private int numSelectionTouchpoints = 0;
    private double touchX, touchY;
    private String name;
    private double lastY = 250.0;
    private double lastX = 250.0;

    private double edgeZoomDelta = 0;
    private double angleDelta = 0;
    private Point2D midPoint = null;
    private float edgeZoom = 1.0f;

    private double selectionZoomDelta = 0;
    private float selectionZoom = 1.0f;

    private boolean twoFinger = false;
    private float selectionHeight;

    private Rectangle touchpad;
    private Rectangle horizontal_pad, vertical_pad;
    private Rectangle debugRect;

    private ArrayList<TouchPoint> edgeTPList;
    private ArrayList<TouchPoint> selectionTPList;
    private BoxModelObject b3d;

    private BoxModelWidget(Builder builder) {
        super();

        this.width = builder.width;
        this.height = builder.height;

        this.touchpad = new Rectangle(this.width, this.height);
        this.touchpad.setTranslateX(100);
        this.getChildren().add(touchpad);

        this.horizontal_pad = new Rectangle(80, this.height);
        this.getChildren().add(horizontal_pad);
        
        this.vertical_pad = new Rectangle(this.width, 80);
        this.vertical_pad.setTranslateY(this.height - 80);
        this.vertical_pad.setTranslateX(-(this.width + 5));
        this.getChildren().add(vertical_pad);

        this.debugRect = new Rectangle(80, 1);
        debugRect.setFill(Color.TRANSPARENT);
        //this.getChildren().add(debugRect);
        this.name = builder.name;

        edgeTPList = new ArrayList<TouchPoint>();
        selectionTPList = new ArrayList<TouchPoint>();

        this.addEventFilter(
                MouseEvent.ANY,
                new EventHandler<MouseEvent>() {
                    public void handle(final MouseEvent mouseEvent) {

                        mouseEvent.consume();

                    }
                });
        this.addEventFilter(
                ZoomEvent.ANY,
                new EventHandler<ZoomEvent>() {
                    public void handle(final ZoomEvent event) {

                        event.consume();

                    }
                });

        // Jump if the first finger touched the ball and is either
        // moving or still, and the second finger touches a rectangle
        EventHandler<TouchEvent> loopScaleHandler = (TouchEvent event) -> {
            
            if (numEdgeTouchpoints != 2 || edgeTPList.size() != 2) {
                // Ignore if this is not a two-finger touch
                return;
            }
            
            //reset deltas
            edgeZoomDelta = 0;
            angleDelta = 0;
            midPoint = null;
            edgeZoom = 1.0f;

            //Retrieve values from old TouchPoint location
            TouchPoint tp1 = edgeTPList.get(0);
            TouchPoint tp2 = edgeTPList.get(1);

            double distanceOld = getDistanceBetweenTwoTouchPoints(tp1, tp2);
            double angleOld = getAnglesBetweenTwoTouchPoints(tp1, tp2);

            //Now clear the list and insert the updated TouchPoints
            edgeTPList.clear();
            
            for (TouchPoint tp : event.getTouchPoints()) {
                if (tp.getId() == tp1.getId() || tp.getId() == tp2.getId()) {
                    edgeTPList.add(tp);
                }
            }
            
            //Calculate new values from current TouchPoint location
            tp1 = edgeTPList.get(0);
            tp2 = edgeTPList.get(1);

            double distance = getDistanceBetweenTwoTouchPoints(tp1, tp2);
            double angle = getAnglesBetweenTwoTouchPoints(tp1, tp2);

            edgeZoomDelta = distance - distanceOld;
            
            //simple Filter to prevent flickering when touch is stationary
            if (Math.abs(edgeZoomDelta) < 1.0) {
                return;
            }
            
            if(Math.abs(edgeZoomDelta) < 2.0){
                edgeZoomDelta /= 2;
            }
            
            if(Math.abs(edgeZoomDelta) >= 10.0){
                edgeZoomDelta *= 2;
            }
            
            angleDelta = angle - angleOld;
            midPoint = getMidpointBetweenTwoTouchPoints(tp1, tp2);
            edgeZoom += edgeZoomDelta / 100;

            

            

            if (this.state == BoxModelWidget.SCALE_LOOPS_H ) {

                //Applying values...
                if (edgeZoomDelta != 0) //b3d.getHorizontalEdgeLoop(mpy, 0.2f,zoom);
                {
                    b3d.setScaleFactor(edgeZoom, true);
                }
            }else if (this.state == BoxModelWidget.SCALE_LOOPS_V ) {

                //Applying values...
                if (edgeZoomDelta != 0) //b3d.getHorizontalEdgeLoop(mpy, 0.2f,zoom);
                {
                    b3d.setScaleFactor(edgeZoom, false);
                }
            } else if (this.state == BoxModelWidget.EXTRUDE) {
                
            }
            event.consume();
        };

        this.touchpad.setOnTouchPressed((TouchEvent event) -> {
            numEdgeTouchpoints = 0;
            edgeTPList.clear();

//            if (selectionTPList.isEmpty()) {
//                setState(BoxModelWidget.EXTRUDE);
//               
//            } else {
//                setState(BoxModelWidget.SCALE_LOOPS_H);
//            }

            for (TouchPoint tp : event.getTouchPoints()) {
                if (tp.belongsTo(touchpad)) {
                    numEdgeTouchpoints++;
                    edgeTPList.add(tp);
                }
            }
            event.consume();
        });

        this.touchpad.setOnTouchReleased((TouchEvent event) -> {
            numEdgeTouchpoints--;
            edgeTPList.remove(edgeTPList.size() - 1);
            event.consume();
        });

        this.touchpad.setOnTouchMoved(loopScaleHandler);

        EventHandler<TouchEvent> loopSelectionHandler;
        loopSelectionHandler = (TouchEvent event) -> {

            selectionZoomDelta = 0;
            selectionZoom = 1.0f;

            if (numSelectionTouchpoints > 2) {
                // Ignore if this is not a two-finger or one-finger touch
                return;
            }

            midPoint = null;
            if (numSelectionTouchpoints == 1 && !twoFinger) {
                TouchPoint tp1 = event.getTouchPoints().get(0);
                double delta = (tp1.getY() - lastY);
                float mtp = mapTouchPointValueToModel((float) tp1.getY(), 0, 500, 1, -1);
                lastY = event.getTouchPoint().getY();

                if(Math.abs(delta) > 10 && Math.abs(delta) < 20){
                delta *= 2; 
                }
                else if(Math.abs(delta) >= 20){
                delta *= 3; 
                }
                
                delta = mapTouchPointValueToModel((float) delta, 0, 500, 0, (float)b3d.getHeight());

                //System.out.println("delta after: " + delta);
                b3d.setSelectionHeight(mtp, (float) delta);
            }

            if (numSelectionTouchpoints == 2) {
                this.twoFinger = true;
                //Retrieve values from old TouchPoint location
                TouchPoint tp1 = selectionTPList.get(0);
                TouchPoint tp2 = selectionTPList.get(1);

                double distanceOld = getDistanceBetweenTwoTouchPoints(tp1, tp2);

                //Now clear the list and insert the updated TouchPoints
                selectionTPList.clear();

                for (TouchPoint tp : event.getTouchPoints()) {
                    if (tp.getId() == tp1.getId() || tp.getId() == tp2.getId()) {
                        selectionTPList.add(tp);
                    }
                }

                //Calculate new values from current TouchPoint location
                tp1 = selectionTPList.get(0);
                tp2 = selectionTPList.get(1);

                double distance = getDistanceBetweenTwoTouchPoints(tp1, tp2);
                float mappedDistance = mapTouchPointValueToModel((float) distance, 0, 500, 0, 2);

                selectionZoomDelta = distance - distanceOld;
                
                if(Math.abs(selectionZoomDelta) > 5)
                    selectionZoomDelta *= 2;
                
                selectionZoom += selectionZoomDelta;

                b3d.setRangeHorizontal((float) selectionZoomDelta / 500);
            }

            event.consume();
        };

        this.horizontal_pad.setOnTouchPressed((TouchEvent event) -> {
            numSelectionTouchpoints = 0;
            selectionTPList.clear();
            lastY = event.getTouchPoint().getY();
            debugRect.setTranslateY(lastY);
            debugRect.setFill(Color.BEIGE);

            if (edgeTPList.isEmpty()) {
                setState(BoxModelWidget.SCALE_LOOPS_H);
            } else {
                setState(BoxModelWidget.EXTRUDE);
            }

            for (TouchPoint tp : event.getTouchPoints()) {
                if (tp.belongsTo(horizontal_pad)) {
                    numSelectionTouchpoints++;
                    selectionTPList.add(tp);
                }
            }
            event.consume();
        });

        this.horizontal_pad.setOnTouchReleased((TouchEvent event) -> {
            numSelectionTouchpoints--;
            if(selectionTPList.size()>0)
                selectionTPList.remove(selectionTPList.size() - 1);
            if (selectionTPList.size() == 0) {
                twoFinger = false;
                setState(BoxModelWidget.START);
            }

            lastY = event.getTouchPoint().getY();
            debugRect.setTranslateY(lastY);
            debugRect.setFill(Color.TRANSPARENT);
            event.consume();
        });

        this.horizontal_pad.setOnTouchMoved(loopSelectionHandler);
        
        EventHandler<TouchEvent> loopSelectionHandler2;
        loopSelectionHandler2 = (TouchEvent event) -> {

            selectionZoomDelta = 0;
            selectionZoom = 1.0f;

            if (numSelectionTouchpoints > 2) {
                // Ignore if this is not a two-finger or one-finger touch
                return;
            }

            midPoint = null;
            if (numSelectionTouchpoints == 1 && !twoFinger) {
                TouchPoint tp1 = event.getTouchPoints().get(0);
                double delta = (tp1.getX() - lastX);
                float mtp = mapTouchPointValueToModel((float) tp1.getX(), 0, 500, 1, -1);
                lastX = event.getTouchPoint().getX();
                
                if(Math.abs(delta) > 10 && Math.abs(delta) < 20){
                delta *= 2; 
                }
                else if(Math.abs(delta) >= 20){
                delta *= 3; 
                }

                delta = mapTouchPointValueToModel((float) delta, 0, 500, 0, (float)b3d.getHeight());

                b3d.setSelectionWidth(mtp, (float) delta);
                
            }

            if (numSelectionTouchpoints == 2) {
                this.twoFinger = true;
                //Retrieve values from old TouchPoint location
                TouchPoint tp1 = selectionTPList.get(0);
                TouchPoint tp2 = selectionTPList.get(1);

                double distanceOld = getDistanceBetweenTwoTouchPoints(tp1, tp2);

                //Now clear the list and insert the updated TouchPoints
                selectionTPList.clear();

                for (TouchPoint tp : event.getTouchPoints()) {
                    if (tp.getId() == tp1.getId() || tp.getId() == tp2.getId()) {
                        selectionTPList.add(tp);
                    }
                }

                //Calculate new values from current TouchPoint location
                tp1 = selectionTPList.get(0);
                tp2 = selectionTPList.get(1);

                double distance = getDistanceBetweenTwoTouchPoints(tp1, tp2);
                float mappedDistance = mapTouchPointValueToModel((float) distance, 0, 500, 0, 2);

                selectionZoomDelta = distance - distanceOld;
                if(Math.abs(selectionZoomDelta) > 5)
                    selectionZoomDelta *= 2;
                selectionZoom += selectionZoomDelta;

                b3d.setRangeVertical((float) selectionZoomDelta / 500);
            }

            event.consume();
        };
        this.vertical_pad.setOnTouchPressed((TouchEvent event) -> {
            numSelectionTouchpoints = 0;
            selectionTPList.clear();
            lastX = event.getTouchPoint().getX();


            if (edgeTPList.isEmpty()) {
                setState(BoxModelWidget.SCALE_LOOPS_V);
            } else {
                setState(BoxModelWidget.EXTRUDE);
            }

            for (TouchPoint tp : event.getTouchPoints()) {
                if (tp.belongsTo(vertical_pad)) {
                    numSelectionTouchpoints++;
                    selectionTPList.add(tp);
                }
            }
            event.consume();
        });

        this.vertical_pad.setOnTouchReleased((TouchEvent event) -> {
            numSelectionTouchpoints--;
            if(selectionTPList.size()>0)
                selectionTPList.remove(selectionTPList.size() - 1);
            if (selectionTPList.size() == 0) {
                twoFinger = false;
                setState(BoxModelWidget.START);
            }

            lastX = event.getTouchPoint().getX();
            event.consume();
        });

        this.vertical_pad.setOnTouchMoved(loopSelectionHandler2);

    }

    private void setState(int state) {

        this.state = state;

        switch (state) {
            case BoxModelWidget.START:
                //System.out.println("Start state");
                if (b3d.getSelectionWidget() != null) {
                    this.b3d.getSelectionWidget().setVisible(false);
                }if (b3d.getSelectionWidget2() != null) {
                    this.b3d.getSelectionWidget2().setVisible(false);
                }
                break;
            case BoxModelWidget.SCALE_LOOPS_H:
                //System.out.println("Scale loops state");
                this.b3d.getSelectionWidget().setVisible(true);
                break;
            case BoxModelWidget.SCALE_LOOPS_V:
                //System.out.println("Scale loops state");
                this.b3d.getSelectionWidget2().setVisible(true);
                break;      
            case BoxModelWidget.EXTRUDE:
                //System.out.println("Extrude state");
                this.b3d.getSelectionWidget().setVisible(false);
                this.b3d.getSelectionWidget2().setVisible(false);
                break;
            case BoxModelWidget.DRAW:
                //System.out.println("Draw state");
                this.b3d.getSelectionWidget().setVisible(false);
                this.b3d.getSelectionWidget2().setVisible(false);
                break;
        }

    }

    private float mapTouchPointValueToModel(float value, float oldLow, float oldHigh, float newLow, float newHigh) {
        return (value - oldLow) / (oldHigh - oldLow) * (newHigh - newLow) + newLow;
    }

    private Point2D getMidpointBetweenTwoTouchPoints(TouchPoint tp1, TouchPoint tp2) {
        double x1 = tp1.getX();
        double y1 = tp1.getY();
        double x2 = tp2.getX();
        double y2 = tp2.getY();

        return new Point2D(
                (x1 + x2) / 2,
                (y1 + y2) / 2
        );
    }

    private double getDistanceBetweenTwoTouchPoints(TouchPoint tp1, TouchPoint tp2) {

        double distance = 0;

        double x1 = tp1.getScreenX();
        double y1 = tp1.getScreenY();
        double x2 = tp2.getScreenX();
        double y2 = tp2.getScreenY();

        distance = Math.sqrt(
                (x1 - x2) * (x1 - x2)
                + (y1 - y2) * (y1 - y2)
        );
        return distance;
    }

    private double getAnglesBetweenTwoTouchPoints(TouchPoint tp1, TouchPoint tp2) {
        double angle = 0;
        double x1 = tp1.getScreenX();
        double y1 = tp1.getScreenY();
        double x2 = tp2.getScreenX();
        double y2 = tp2.getScreenY();

        double deltaY = y2 - y1;
        double deltaX = x2 - x1;

        angle = Math.atan2(deltaY, deltaX) * 180.0 / Math.PI;

        return angle;
    }

    public void setB3D(BoxModelObject b3d) {
        this.b3d = b3d;
    }
    
    public BoxModelObject getB3D() {
        return b3d;
    }

    public static class Builder {

        private final float width;
        private final float height;
        private final float x;
        private final float y;
        private final String name;

        public Builder(float x, float y, float width, float height, String name) {
            this.width = width;
            this.height = height;
            this.x = x;
            this.y = y;
            this.name = name;

        }

        public BoxModelWidget build() {
            return new BoxModelWidget(this);
        }
    }

}
