/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package pkg3dtestapplication;

import java.util.ArrayList;
import javafx.event.EventHandler;
import javafx.geometry.Point2D;
import javafx.scene.input.RotateEvent;
import javafx.scene.input.TouchEvent;
import javafx.scene.input.TouchPoint;
import javafx.scene.input.ZoomEvent;
import javafx.scene.shape.Rectangle;

/**
 *
 * @author henri
 */
public class TouchControlWidget extends Rectangle {

    private float x, y, width, height;
    private int type;
    private int numTouchpoints = 0;
    private long touchID = -1;
    private double touchX, touchY;
    private String name;

    private double zoomDelta = 0;
    private double angleDelta = 0;
    private Point2D midPoint = null;

    private ArrayList<TouchPoint> tplist;
    private Xform xform;

    public static final int TWO_FINGER_ROTATE = 0;
    public static final int TWO_FINGER_ZOOM = 1;
    public static final int TWO_FINGER_RANGE = 2;

    private TouchControlWidget(Builder builder) {
        super(builder.width, builder.height);
        this.width = builder.width;
        this.height = builder.height;
        setTranslateX(x);
        setTranslateY(y);

        this.xform = builder.xform;
        this.name = builder.name;

        this.type = builder.type;

        tplist = new ArrayList<TouchPoint>();

        // Jump if the first finger touched the ball and is either
        // moving or still, and the second finger touches a rectangle
        EventHandler<TouchEvent> handler = (TouchEvent event) -> {
            
            if (numTouchpoints != 2) {
                // Ignore if this is not a two-finger touch
                return;
            }
            //reset deltas
            zoomDelta = 0;
            angleDelta = 0;
            midPoint = null;
            
            //Retrieve values from old TouchPoint location
            TouchPoint tp1 = tplist.get(0);
            TouchPoint tp2 = tplist.get(1);

            double distanceOld = getDistanceBetweenTwoTouchPoints(tp1, tp2); 
            double angleOld = getAnglesBetweenTwoTouchPoints(tp1, tp2);
            
            //Now clear the list and insert the updated TouchPoints
            tplist.clear();

            for (TouchPoint tp : event.getTouchPoints()) {
                if (tp.getId() == tp1.getId() || tp.getId() == tp2.getId()) {
                    tplist.add(tp);
                }
            }
               
            //Calculate new values from current TouchPoint location
            tp1 = tplist.get(0);
            tp2 = tplist.get(1);
            
            double distance = getDistanceBetweenTwoTouchPoints(tp1, tp2);
            double angle = getAnglesBetweenTwoTouchPoints(tp1, tp2);
            
            zoomDelta = distance - distanceOld;
            angleDelta = angle - angleOld;
            midPoint = getMidpointBetweenTwoTouchPoints(tp1, tp2);
            
            //simple Filter to prevent flickering when touch is stationary
            if (Math.abs(zoomDelta) < 1.0) {
                return;
            }
            
            //Applying values...
            if (this.type == TouchControlWidget.TWO_FINGER_ZOOM) {
                xform.setRotateY(xform.ry.getAngle() + angleDelta);
            }

            if (this.type == TouchControlWidget.TWO_FINGER_ROTATE) {
                xform.setRotateZ(xform.rz.getAngle() + angleDelta);
            }
            
            event.consume();
        };

        setOnTouchPressed((TouchEvent event) -> {
            numTouchpoints = 0;
            tplist.clear();
            for (TouchPoint tp : event.getTouchPoints()) {
                if (tp.belongsTo(this)) {
                    numTouchpoints++;
                    tplist.add(tp);
                }
            }

            touchID = event.getEventSetId();
            System.out.println(name + " touchcount: " + numTouchpoints);
            event.consume();
        });

        setOnTouchReleased((TouchEvent event) -> {
            numTouchpoints--;
            tplist.remove(tplist.size() - 1);
            System.out.println(name + " touchcount: " + numTouchpoints);
            event.consume();
        });

        setOnTouchMoved(handler);

    }
    
    private Point2D getMidpointBetweenTwoTouchPoints(TouchPoint tp1, TouchPoint tp2){
        double x1 = tp1.getScreenX();
        double y1 = tp1.getScreenY();
        double x2 = tp2.getScreenX();
        double y2 = tp2.getScreenY();
        
        return new Point2D(
                (x1+x2)/2,
                (y1+y2/2)
        );
    }

    private double getDistanceBetweenTwoTouchPoints(TouchPoint tp1, TouchPoint tp2) {
        
        double distance = 0;
        
        double x1 = tp1.getScreenX();
        double y1 = tp1.getScreenY();
        double x2 = tp2.getScreenX();
        double y2 = tp2.getScreenY();

        distance = Math.sqrt(
                (x1 - x2) * (x1 - x2)
                + (y1 - y2) * (y1 - y2)
        );
        return distance;
    }
    
    private double getAnglesBetweenTwoTouchPoints(TouchPoint tp1, TouchPoint tp2){
        double angle = 0;
        double x1 = tp1.getScreenX();
        double y1 = tp1.getScreenY();
        double x2 = tp2.getScreenX();
        double y2 = tp2.getScreenY();
        
        double deltaY = y2 - y1;
        double deltaX = x2 - x1;
        
        angle = Math.atan2(deltaY, deltaX) * 180.0 / Math.PI;
        
        return angle;
    }

    public static class Builder {

        private final float width;
        private final float height;
        private final float x;
        private final float y;
        private final Xform xform;
        private final String name;
        private final int type;

        public Builder(float x, float y, float width, float height, Xform xform, String name, int type) {
            this.width = width;
            this.height = height;
            this.x = x;
            this.y = y;
            this.xform = xform;
            this.name = name;
            this.type = type;
        }

        public TouchControlWidget build() {
            return new TouchControlWidget(this);
        }
    }

}
